import os
import math
import numpy as np
import random
from tqdm import tqdm
from tqdm.auto import trange
import yaml
import datetime
import argparse
from shutil import copy2

import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import torch.nn.functional as F

from models.encoder import ResnetEncoder
from models.decoder import DepthDecoder_v1, DepthDecoder_v2
from models.rnn import ConvLSTMModel_v1, ConvLSTMModel_v2, ConvGRUModel_v1, ConvGRUModel_v2

from data_load import KITTIDataset_v1, KITTIDataset_v2, KITTIDataset_v3
from loss import MaskedMSELoss, MaskedL1Loss, BerHu, get_smooth_loss

from utils import MetricAggregator, count_scene_frames_func, generate_frame_seq_func, get_mask_func, get_context_vectors, get_context_vector
from utils import disp_to_depth, transform_depth, sort_test_seq_func
 

def forward_step_rnn_v2(rgb_seq, encoder, decoder, clstm_model, config, device):

    n = config['n_sequence']
    k = config['att_n_mask']
    bs = rgb_seq.shape[0]

    feature_arr = []
    encoder_output = []

    for i in range(n+k):
        lvl1, lvl2, lvl3, lvl4, lvl5 = encoder(rgb_seq[:,i])
        if i >= k:
            feature_arr.append([lvl1, lvl2, lvl3, lvl4])
        encoder_output.append(lvl5)
            
    encoder_output = torch.stack(encoder_output, 1)
    exp_v = config['exp_v']
    
    # clstm part

    # initialize h0, c0 for both layers

    if bs > 1: 
        hidden_state_1 = (clstm_model.h0_layer1.repeat(bs, 1, 1, 1), clstm_model.c0_layer1.repeat(bs, 1, 1, 1))
        hidden_state_2 = (clstm_model.h0_layer2.repeat(bs, 1, 1, 1), clstm_model.c0_layer2.repeat(bs, 1, 1, 1))
    else:
        hidden_state_1 = (clstm_model.h0_layer1, clstm_model.c0_layer1)
        hidden_state_2 = (clstm_model.h0_layer2, clstm_model.c0_layer2)

    # iterate through sequence
    lstm_1_output = []
    lstm_2_output = []

    for i in range(n+k):
        hidden_state_1, hidden_state_2 = clstm_model(encoder_output[:,i], hidden_state_1, hidden_state_2)
        lstm_1_output.append(hidden_state_1[0])
        lstm_2_output.append(hidden_state_2[0])

    lstm_1_output = torch.stack(lstm_1_output, 1)
    lstm_2_output = torch.stack(lstm_2_output, 1)

    # attention step

    context_vector = get_context_vectors(lstm_1_output, lstm_2_output, n, k, device)    

    lstm_output = torch.cat((lstm_2_output[:,k:,:], context_vector), 2)
    # concating encoder output + rnn output
    #lstm_output = lstm_2_output

    # decoder part
    depth_output = []

    for i in range(n):
        input_features = feature_arr[i] + [lstm_output[:,i]]

        disp = decoder(input_features)

        _, depth = disp_to_depth(disp)

        depth_output.append(depth)
            
    return torch.stack(depth_output, 1)    

def forward_step_rnn_v1(rgb_seq, encoder, decoder, clstm_model, config, device):

    n = config['n_sequence']
    k = config['att_n_mask']
    bs = rgb_seq.shape[0]
    feature_arr = []
    encoder_output = []

    for i in range(n+k):
        lvl1, lvl2, lvl3, lvl4, lvl5 = encoder(rgb_seq[:,i])
        if i >= k:
            feature_arr.append([lvl1, lvl2, lvl3, lvl4])
        encoder_output.append(lvl5)
            
    encoder_output = torch.stack(encoder_output, 1)
    
    # clstm part

    # initialize h0, c0

    if bs > 1: 
        hidden_state_1 = (clstm_model.h0_layer1.repeat(bs, 1, 1, 1), clstm_model.c0_layer1.repeat(bs, 1, 1, 1))
    else:
        hidden_state_1 = (clstm_model.h0_layer1, clstm_model.c0_layer1)

    # iterate through sequence
    lstm_output = []

    for i in range(n+k):
        hidden_state_1 = clstm_model(encoder_output[:,i], hidden_state_1)
        lstm_output.append(hidden_state_1[0])

    lstm_output = torch.stack(lstm_output, 1)  

    # decoder part
    lstm_output = lstm_output[:,k:,:]

    depth_output = []

    for i in range(n):
        input_features = feature_arr[i] + [lstm_output[:,i]]

        disp = decoder(input_features)

        _, depth = disp_to_depth(disp)

        depth_output.append(depth)
            
    return torch.stack(depth_output, 1)   

def forward_step_rnn_v3(rgb_seq, encoder, decoder, cgru_model, config, device):

    n = config['n_sequence']
    k = config['att_n_mask']
    bs = rgb_seq.shape[0]
    feature_arr = []
    encoder_output = []

    for i in range(n+k):
        lvl1, lvl2, lvl3, lvl4, lvl5 = encoder(rgb_seq[:,i])
        if i >= k:
            feature_arr.append([lvl1, lvl2, lvl3, lvl4])
        encoder_output.append(lvl5)
            
    encoder_output = torch.stack(encoder_output, 1)
    
    # clstm part

    # initialize h0, c0

    if bs > 1: 
        hidden_state_1 = cgru_model.h0_layer1.repeat(bs, 1, 1, 1)
    else:
        hidden_state_1 = cgru_model.h0_layer1

    # hidden_state_1 = torch.zeros(bs, cgru_model.hidden_dim, cgru_model.height, cgru_model.width, device=device)

    # iterate through sequence
    gru_output = []

    for i in range(n+k):
        hidden_state_1 = cgru_model(encoder_output[:,i], hidden_state_1)
        gru_output.append(hidden_state_1)

    gru_output = torch.stack(gru_output, 1)  

    # decoder part
    gru_output = gru_output[:,k:,:]

    depth_output = []

    for i in range(n):
        input_features = feature_arr[i] + [gru_output[:,i]]

        disp = decoder(input_features)

        _, depth = disp_to_depth(disp)

        depth_output.append(depth)
            
    return torch.stack(depth_output, 1)   

def forward_step_rnn_v4(rgb_seq, encoder, decoder, cgru_model, config, device):

    n = config['n_sequence']
    k = config['att_n_mask']
    bs = rgb_seq.shape[0]
    feature_arr = []
    encoder_output = []

    for i in range(n+k):
        lvl1, lvl2, lvl3, lvl4, lvl5 = encoder(rgb_seq[:,i])
        if i >= k:
            feature_arr.append([lvl1, lvl2, lvl3, lvl4])
        encoder_output.append(lvl5)
            
    encoder_output = torch.stack(encoder_output, 1)
    
    # clstm part

    # initialize h0, c0

    if bs > 1: 
        hidden_state_1 = cgru_model.h0_layer1.repeat(bs, 1, 1, 1)
    else:
        hidden_state_1 = cgru_model.h0_layer1

    # hidden_state_1 = torch.zeros(bs, cgru_model.hidden_dim, cgru_model.height, cgru_model.width, device=device)

    # iterate through sequence
    gru_output = []

    for i in range(n+k):
        hidden_state_1 = cgru_model(encoder_output[:,i], hidden_state_1)
        gru_output.append(hidden_state_1)

    gru_output = torch.stack(gru_output, 1)  

    # decoder part
    #gru_output = gru_output[:,k:,:]

    # either stack previous and current or fuse features
    if config['v4_type'] == 'concat':
        output = torch.cat((gru_output[:,1:,:], gru_output[:,:-1,:]), 2)
    elif config['v4_type'] == 'fuse':
        output = (gru_output[:,1:,:] + gru_output[:,:-1,:]) / 2

    depth_output = []

    for i in range(n):
        input_features = feature_arr[i] + [output[:,i]]

        disp = decoder(input_features)

        _, depth = disp_to_depth(disp)

        depth_output.append(depth)
            
    return torch.stack(depth_output, 1)   

def forward_step_rnn_v5(rgb_seq, encoder, decoder, clstm_model, config, device):

    n = config['n_sequence']
    k = config['att_n_mask']
    bs = rgb_seq.shape[0]
    feature_arr = []
    encoder_output = []

    for i in range(n+k):
        lvl1, lvl2, lvl3, lvl4, lvl5 = encoder(rgb_seq[:,i])
        if i >= k:
            feature_arr.append([lvl1, lvl2, lvl3, lvl4])
        encoder_output.append(lvl5)
            
    encoder_output = torch.stack(encoder_output, 1)
    
    # clstm part

    # initialize h0, c0

    if bs > 1: 
        hidden_state_1 = (clstm_model.h0_layer1.repeat(bs, 1, 1, 1), clstm_model.c0_layer1.repeat(bs, 1, 1, 1))
    else:
        hidden_state_1 = (clstm_model.h0_layer1, clstm_model.c0_layer1)

    # hidden_state_1 = torch.zeros(bs, cgru_model.hidden_dim, cgru_model.height, cgru_model.width, device=device)

    # iterate through sequence
    lstm_output = []

    for i in range(n+k):
        hidden_state_1 = clstm_model(encoder_output[:,i], hidden_state_1)
        lstm_output.append(hidden_state_1[0])

    lstm_output = torch.stack(lstm_output, 1)  

    # decoder part

    # either stack previous and current or fuse features
    if config['v5_type'] == 'concat':
        output = torch.cat((lstm_output[:,1:,:], lstm_output[:,:-1,:]), 2)
    elif config['v5_type'] == 'fuse':
        output = (lstm_output[:,1:,:] + lstm_output[:,:-1,:]) / 2

    depth_output = []

    for i in range(n):
        input_features = feature_arr[i] + [output[:,i]]

        disp = decoder(input_features)

        _, depth = disp_to_depth(disp)

        depth_output.append(depth)
            
    return torch.stack(depth_output, 1)   

def forward_step_rnn_v6(rgb_seq, encoder, decoder, cgru_model, config, device):

    n = config['n_sequence']
    k = config['att_n_mask']
    bs = rgb_seq.shape[0]

    feature_arr = []
    encoder_output = []

    for i in range(n+k):
        lvl1, lvl2, lvl3, lvl4, lvl5 = encoder(rgb_seq[:,i])
        if i >= k:
            feature_arr.append([lvl1, lvl2, lvl3, lvl4])
        encoder_output.append(lvl5)
            
    encoder_output = torch.stack(encoder_output, 1)
    exp_v = config['exp_v']
    
    # clstm part

    # initialize h0, c0 for both layers

    if bs > 1: 
        hidden_state_1 = cgru_model.h0_layer1.repeat(bs, 1, 1, 1)
        hidden_state_2 = cgru_model.h0_layer2.repeat(bs, 1, 1, 1)
    else:
        hidden_state_1 = cgru_model.h0_layer1
        hidden_state_2 = cgru_model.h0_layer2

    # iterate through sequence
    gru_1_output = []
    gru_2_output = []

    for i in range(n+k):
        hidden_state_1, hidden_state_2 = cgru_model(encoder_output[:,i], hidden_state_1, hidden_state_2)
        gru_1_output.append(hidden_state_1)
        gru_2_output.append(hidden_state_2)

    gru_1_output = torch.stack(gru_1_output, 1)
    gru_2_output = torch.stack(gru_2_output, 1)

    # attention step

    context_vector = get_context_vectors(gru_1_output, gru_2_output, n, k, device)    

    gru_output = torch.cat((gru_2_output[:,k:,:], context_vector), 2)
    # concating encoder output + rnn output
    #lstm_output = lstm_2_output

    # decoder part
    depth_output = []

    for i in range(n):
        input_features = feature_arr[i] + [gru_output[:,i]]

        disp = decoder(input_features)

        _, depth = disp_to_depth(disp)

        depth_output.append(depth)
            
    return torch.stack(depth_output, 1)    

def test_validate_v1(test_frames_dict, encoder, clstm, decoder, config, log_file):
    
    metric_aggregator = MetricAggregator()
    sequence_names = list(test_frames_dict.keys())
    with torch.no_grad():
        for j in trange(len(sequence_names)):
            sequence_name = sequence_names[j]
            frame_list = test_frames_dict[sequence_name]

            frame_count = 0
            dataset = KITTIDataset_v3(config, frame_list, sequence_name)
            
            # initialize zeros at the beginning or initialize 
            #hidden_state = clstm.init_hidden(clstm.hidden_dim, clstm.height, clstm.width, config['device'])
            if config['exp_v'] in ['v1','v5']:
                hidden_state = (clstm.h0_layer1, clstm.c0_layer1)
            elif config['exp_v'] in ['v3','v4']:
                hidden_state = clstm.h0_layer1

            # iterate through sequence, until all frames from test files for corresponding scene are all seen
            for i in range(len(dataset)):
            #for i in trange(len(dataset), desc=sequence_name.split('/')[1]):
            #for i in tqdm(range(len(dataset))):
                
                frame_idx, rgb, depth_gt = dataset[i]
                rgb = rgb.to(config['device'])
                
                feature_arr = encoder(rgb)                    
                
                new_hidden_state = clstm(feature_arr[-1], hidden_state)                

                # if we are on file, which needs evaluation, put it through decoder and evaluate it
                if frame_idx in frame_list:
                    depth_gt = depth_gt.to(config['device'])

                    if config['exp_v'] == 'v1':
                        output = new_hidden_state[0]
                    elif config['exp_v'] == 'v3':
                        output = new_hidden_state
                    elif config['exp_v'] == 'v4':
                        if i == 0:
                            prev_state = new_hidden_state
                        else:
                            prev_state = hidden_state

                        if config['v4_type'] == 'concat':
                            output = torch.cat((new_hidden_state,prev_state), 1)
                        elif config['v4_type'] == 'fuse':
                            output = (new_hidden_state + prev_state) / 2
                    elif config['exp_v'] == 'v5':
                        if i == 0:
                            prev_state = new_hidden_state
                        else:
                            prev_state = hidden_state

                        if config['v5_type'] == 'concat':
                            output = torch.cat((new_hidden_state[0],prev_state[0]), 1)
                        elif config['v5_type'] == 'fuse':
                            output = (new_hidden_state[0] + prev_state[0]) / 2

                    #print(output.shape)
                    disp = decoder(feature_arr[:-1] + [output])
                    _, depth_pred = disp_to_depth(disp)
                    depth_pred = F.interpolate(depth_pred, 
                                               [config['true_height'], config['true_width']], 
                                               mode="bilinear", 
                                               align_corners=False)
                    
                    depth_pred = torch.clamp(depth_pred, 1e-3, 80)
                    depth_pred, depth_gt = transform_depth(depth_pred, depth_gt)
                    metric_aggregator.evaluate(depth_pred.data, depth_gt.data)
                    
                    # store info, that we passed one file
                    frame_count += 1
                    if frame_count == len(frame_list):
                        break

                hidden_state = new_hidden_state
        
        print('-------------')
        print("Finished with all sequences")
        print('-------------')
        metric_dict = metric_aggregator.average()
        
        msg_1 = "Validation on test samples:"
        msg_2 = f"RMSE: {metric_dict['rmse']:.3f} RMSE_log: {metric_dict['rmse_log']:.3f} REL: {metric_dict['absrel']:.3f} Sq_rel: {metric_dict['sq_rel']:.3f}"
        msg_3 = f"Delta1: {metric_dict['delta1']:.3f} Delta2: {metric_dict['delta2']:.3f} Delta3: {metric_dict['delta3']:.3f}" 
        print(msg_1)
        print(msg_2)
        print(msg_3)

        if log_file:
            with open(log_file, 'a') as f:
                print(msg_1, file=f)
                print(msg_2, file=f)
                print(msg_3, file=f)

def test_validate_v2(test_frames_dict, encoder, clstm, decoder, config, log_file):
    
    metric_aggregator = MetricAggregator()
    sequence_names = list(test_frames_dict.keys())
    with torch.no_grad():
        for j in trange(len(sequence_names)):
            sequence_name = sequence_names[j]
            frame_list = test_frames_dict[sequence_name]

            frame_count = 0
            dataset = KITTIDataset_v3(config, frame_list, sequence_name)
            
            # initialize zeros at the beginning or initialize 
            #hidden_state = clstm.init_hidden(clstm.hidden_dim, clstm.height, clstm.width, config['device'])
            if config['exp_v'] == 'v2':
                hidden_state_1 = (clstm.h0_layer1, clstm.c0_layer1)
                hidden_state_2 = (clstm.h0_layer2, clstm.c0_layer2)

            # iterate through sequence, until all frames from test files for corresponding scene are all seen
            for i in range(len(dataset)):
            #for i in trange(len(dataset), desc=sequence_name.split('/')[1]):
            #for i in tqdm(range(len(dataset))):
                
                frame_idx, rgb, depth_gt = dataset[i]
                rgb = rgb.to(config['device'])
                
                feature_arr = encoder(rgb)                    
                
                new_hidden_state_1, new_hidden_state_2 = clstm(feature_arr[-1], hidden_state_1, hidden_state_2)          

                if i == 0:
                    prev_states_1 = [new_hidden_state_1[0].view(1, -1)]*config['att_n_mask']
                    prev_states_2 = [new_hidden_state_2[0].view(1, -1)]*config['att_n_mask']


                l2_shape = new_hidden_state_2[0].shape
                # if we are on file, which needs evaluation, put it through decoder and evaluate it
                if frame_idx in frame_list:
                    depth_gt = depth_gt.to(config['device'])

                    context_vector = get_context_vector(new_hidden_state_1[0], prev_states_1, prev_states_2, l2_shape)

                    output = torch.cat((new_hidden_state_2[0], context_vector), 1)
                    #print(output.shape, context_vector.shape)
                    #print(output.shape)
                    disp = decoder(feature_arr[:-1] + [output])
                    _, depth_pred = disp_to_depth(disp)
                    depth_pred = F.interpolate(depth_pred, 
                                               [config['true_height'], config['true_width']], 
                                               mode="bilinear", 
                                               align_corners=False)
                    
                    depth_pred = torch.clamp(depth_pred, 1e-3, 80)
                    depth_pred, depth_gt = transform_depth(depth_pred, depth_gt)
                    metric_aggregator.evaluate(depth_pred.data, depth_gt.data)
                    
                    # store info, that we passed one file
                    frame_count += 1
                    if frame_count == len(frame_list):
                        break

                hidden_state_1, hidden_state_2 = new_hidden_state_1, new_hidden_state_2

                if i > 0:
                    prev_states_1.pop(0)
                    prev_states_2.pop(0)
                    prev_states_1.append(new_hidden_state_1[0].view(1, -1))
                    prev_states_2.append(new_hidden_state_2[0].view(1, -1))

        
        print('-------------')
        print("Finished with all sequences")
        print('-------------')
        metric_dict = metric_aggregator.average()
        
        msg_1 = "Validation on test samples:"
        msg_2 = f"RMSE: {metric_dict['rmse']:.3f} RMSE_log: {metric_dict['rmse_log']:.3f} REL: {metric_dict['absrel']:.3f} Sq_rel: {metric_dict['sq_rel']:.3f}"
        msg_3 = f"Delta1: {metric_dict['delta1']:.3f} Delta2: {metric_dict['delta2']:.3f} Delta3: {metric_dict['delta3']:.3f}" 
        print(msg_1)
        print(msg_2)
        print(msg_3)

        if log_file:
            with open(log_file, 'a') as f:
                print(msg_1, file=f)
                print(msg_2, file=f)
                print(msg_3, file=f)

def test_validate_v3(test_frames_dict, encoder, cgru, decoder, config, log_file):
    
    metric_aggregator = MetricAggregator()
    sequence_names = list(test_frames_dict.keys())
    with torch.no_grad():
        for j in trange(len(sequence_names)):
            sequence_name = sequence_names[j]
            frame_list = test_frames_dict[sequence_name]

            frame_count = 0
            dataset = KITTIDataset_v3(config, frame_list, sequence_name)
            
            # initialize zeros at the beginning or initialize 
            #hidden_state = clstm.init_hidden(clstm.hidden_dim, clstm.height, clstm.width, config['device'])
            if config['exp_v'] == 'v6':
                hidden_state_1 = cgru.h0_layer1
                hidden_state_2 = cgru.h0_layer2

            # iterate through sequence, until all frames from test files for corresponding scene are all seen
            for i in range(len(dataset)):
            #for i in trange(len(dataset), desc=sequence_name.split('/')[1]):
            #for i in tqdm(range(len(dataset))):
                
                frame_idx, rgb, depth_gt = dataset[i]
                rgb = rgb.to(config['device'])
                
                feature_arr = encoder(rgb)                    
                
                new_hidden_state_1, new_hidden_state_2 = cgru(feature_arr[-1], hidden_state_1, hidden_state_2)          

                if i == 0:
                    prev_states_1 = [new_hidden_state_1.view(1, -1)]*config['att_n_mask']
                    prev_states_2 = [new_hidden_state_2.view(1, -1)]*config['att_n_mask']


                l2_shape = new_hidden_state_2.shape
                # if we are on file, which needs evaluation, put it through decoder and evaluate it
                if frame_idx in frame_list:
                    depth_gt = depth_gt.to(config['device'])

                    context_vector = get_context_vector(new_hidden_state_1, prev_states_1, prev_states_2, l2_shape)

                    output = torch.cat((new_hidden_state_2, context_vector), 1)
                    #print(output.shape, context_vector.shape)
                    #print(output.shape)
                    disp = decoder(feature_arr[:-1] + [output])
                    _, depth_pred = disp_to_depth(disp)
                    depth_pred = F.interpolate(depth_pred, 
                                               [config['true_height'], config['true_width']], 
                                               mode="bilinear", 
                                               align_corners=False)
                    
                    depth_pred = torch.clamp(depth_pred, 1e-3, 80)
                    depth_pred, depth_gt = transform_depth(depth_pred, depth_gt)
                    metric_aggregator.evaluate(depth_pred.data, depth_gt.data)
                    
                    # store info, that we passed one file
                    frame_count += 1
                    if frame_count == len(frame_list):
                        break

                hidden_state_1, hidden_state_2 = new_hidden_state_1, new_hidden_state_2

                if i > 0:
                    prev_states_1.pop(0)
                    prev_states_2.pop(0)
                    prev_states_1.append(new_hidden_state_1.view(1, -1))
                    prev_states_2.append(new_hidden_state_2.view(1, -1))

        print('-------------')
        print("Finished with all sequences")
        print('-------------')
        metric_dict = metric_aggregator.average()
        
        msg_1 = "Validation on test samples:"
        msg_2 = f"RMSE: {metric_dict['rmse']:.3f} RMSE_log: {metric_dict['rmse_log']:.3f} REL: {metric_dict['absrel']:.3f} Sq_rel: {metric_dict['sq_rel']:.3f}"
        msg_3 = f"Delta1: {metric_dict['delta1']:.3f} Delta2: {metric_dict['delta2']:.3f} Delta3: {metric_dict['delta3']:.3f}" 
        print(msg_1)
        print(msg_2)
        print(msg_3)

        if log_file:
            with open(log_file, 'a') as f:
                print(msg_1, file=f)
                print(msg_2, file=f)
                print(msg_3, file=f)


def train_epoch_rnn(train_loader, encoder, decoder, clstm_model, optimizer, epoch, device, config, log_file):
    
    metric_aggregator = MetricAggregator()
    k = config['att_n_mask']
    exp_var = config['exp_v']
    t = tqdm(train_loader)
    #running_loss = 0.0

    for i, data in enumerate(t):

        rgb, depth_gt = data
        rgb = rgb.to(device)
        depth_gt = depth_gt.to(device)
        # forward pass
        if exp_var == 'v1':
            depth_pred = forward_step_rnn_v1(rgb, encoder, decoder, clstm_model, config, device)
        elif exp_var == 'v2':
            depth_pred = forward_step_rnn_v2(rgb, encoder, decoder, clstm_model, config, device)
        elif exp_var == 'v3':
            depth_pred = forward_step_rnn_v3(rgb, encoder, decoder, clstm_model, config, device)
        elif exp_var == 'v4':
            depth_pred = forward_step_rnn_v4(rgb, encoder, decoder, clstm_model, config, device)
        elif exp_var == 'v5':
            depth_pred = forward_step_rnn_v5(rgb, encoder, decoder, clstm_model, config, device)
        elif exp_var == 'v6':
            depth_pred = forward_step_rnn_v6(rgb, encoder, decoder, clstm_model, config, device)
        # loss part
        optimizer.zero_grad()

        loss = 0 
        # smooth_loss
        depth_pred = depth_pred.reshape(-1, 1, config['height'], config['width'])

        rgb = rgb[:,k:,:].reshape(-1, 3, config['height'], config['width'])

        if config['use_smooth_loss']:            
            mean_disp = depth_pred.mean(2, True).mean(3, True)
            norm_disp = depth_pred / (mean_disp + 1e-7)
            loss += config['smooth_loss_weight'] * get_smooth_loss(norm_disp, rgb)

        depth_gt = depth_gt.reshape(-1, 1, config['true_height'], config['true_width'])

        depth_pred = torch.clamp(F.interpolate(depth_pred, [config['true_height'], config['true_width']], mode="bilinear", align_corners=False), 1e-3, 80)
        
        loss += config['main_loss_weight'] * criterion(depth_pred, depth_gt)

        loss.backward()

        #running_loss += loss.item()

        optimizer.step()

        # rescale with median
        depth_pred, depth_gt = transform_depth(depth_pred, depth_gt)

        # measure accuracy and record loss
        #metric_aggregator.evaluate(depth_pred.data, depth_gt.data)
        t.set_description("Loss: {}".format(loss.item()))

        # if (i + 1) % config['log_interval'] == 0:
        #     metric_dict = metric_aggregator.average()

        #     avg_loss_value = running_loss / config['log_interval']

        #     msg = f"[{epoch+1}, {i+1}] Loss: {avg_loss_value:.3f} RMSE: {metric_dict['rmse']:.3f}"
        #     print(msg)

        #     with open(log_file, 'a') as f:
        #         print(msg, file=f)

        #     metric_aggregator.reset()
        #     running_loss = 0.0

def validate_rnn(val_loader, encoder, decoder, clstm_model, device, epoch, log_file):
    metric_aggregator = MetricAggregator()

    t = tqdm(val_loader)
    exp_var = config['exp_v']

    for i, data in enumerate(t):
        rgb, depth_gt = data
        rgb = rgb.to(device)
        depth_gt = depth_gt.to(device)
        
        # forward pass
        with torch.no_grad():
            if exp_var == 'v1':
                depth_pred = forward_step_rnn_v1(rgb, encoder, decoder, clstm_model, config, device)
            elif exp_var == 'v2':
                depth_pred = forward_step_rnn_v2(rgb, encoder, decoder, clstm_model, config, device)
            elif exp_var == 'v3':
                depth_pred = forward_step_rnn_v3(rgb, encoder, decoder, clstm_model, config, device)
            elif exp_var == 'v4':
                depth_pred = forward_step_rnn_v4(rgb, encoder, decoder, clstm_model, config, device)
            
        depth_pred = depth_pred.view(-1, 1, config['height'], config['width'])
        depth_pred = torch.clamp(F.interpolate(depth_pred, [config['true_height'], config['true_width']], mode="bilinear", align_corners=False), 1e-3, 80)
        
        depth_gt = depth_gt.view(-1, 1, config['true_height'], config['true_width'])

        depth_pred, depth_gt = transform_depth(depth_pred, depth_gt)

        # measure accuracy and record loss
        metric_aggregator.evaluate(depth_pred.data, depth_gt.data)


    metric_dict = metric_aggregator.average()

    msg_1 = "Epoch: {}, Validation on test samples:".format(epoch)
    msg_2 = f"RMSE: {metric_dict['rmse']:.3f} RMSE_log: {metric_dict['rmse_log']:.3f} REL: {metric_dict['absrel']:.3f} Sq_rel: {metric_dict['sq_rel']:.3f}"
    msg_3 = f"Delta1: {metric_dict['delta1']:.3f} Delta2: {metric_dict['delta2']:.3f} Delta3: {metric_dict['delta3']:.3f}" 
    print(msg_1)
    print(msg_2)
    print(msg_3)

    with open(log_file, 'a') as f:
        print(msg_1, file=f)
        print(msg_2, file=f)
        print(msg_3, file=f)


def forward_step(rgb, encoder, decoder, config, device):

    # encoder
    feature_arr = encoder(rgb)
    # decoder to disp
    disp = decoder(feature_arr)

    # transform disp to depth
    _, depth = disp_to_depth(disp)

    return depth 

def train_epoch(train_loader, encoder, decoder, optimizer, epoch, device, config, log_file):
    
    metric_aggregator = MetricAggregator()
    t = tqdm(train_loader)
    running_loss = 0.0

    for i, data in enumerate(t):
        rgb, depth_gt = data
        rgb = rgb.to(device)
        depth_gt = depth_gt.to(device)
        
        # forward pass

        depth_pred = forward_step(rgb, encoder, decoder, config, device)

        # loss part
        optimizer.zero_grad()

        loss = 0 

        if config['use_smooth_loss']:            
            mean_disp = depth_pred.mean(2, True).mean(3, True)
            norm_disp = depth_pred / (mean_disp + 1e-7)
            loss += config['smooth_loss_weight'] * get_smooth_loss(norm_disp, rgb)

        depth_pred = torch.clamp(F.interpolate(depth_pred, [config['true_height'], config['true_width']], mode="bilinear", align_corners=False), 1e-3, 80)

        loss += config['main_loss_weight'] * criterion(depth_pred, depth_gt)

        loss.backward()

        running_loss += loss.item()

        optimizer.step()

        #evaluate metircs
    
        depth_pred, depth_gt = transform_depth(depth_pred, depth_gt)

        # measure accuracy and record loss
        metric_aggregator.evaluate(depth_pred.data, depth_gt.data)

        if (i + 1) % config['log_interval'] == 0:
            metric_dict = metric_aggregator.average()

            avg_loss_value = running_loss / config['log_interval']

            msg = f"[{epoch+1}, {i+1}] Loss: {avg_loss_value:.3f} RMSE: {metric_dict['rmse']:.3f}"
            print(msg)

            with open(log_file, 'a') as f:
                print(msg, file=f)

            metric_aggregator.reset()
            running_loss = 0.0

def validate(val_loader, encoder, decoder, device, epoch, log_file):
    metric_aggregator = MetricAggregator()

    t = tqdm(val_loader)

    for i, data in enumerate(t):
        rgb, depth_gt = data
        rgb = rgb.to(device)
        depth_gt = depth_gt.to(device)
        
        # forward pass

        with torch.no_grad():
            depth_pred = forward_step(rgb, encoder, decoder, config, device)
 
        depth_pred = torch.clamp(F.interpolate(depth_pred, [config['true_height'], config['true_width']], mode="bilinear", align_corners=False), 1e-3, 80)

        depth_pred, depth_gt = transform_depth(depth_pred, depth_gt)
        # measure accuracy and record loss
        metric_aggregator.evaluate(depth_pred.data, depth_gt.data)

    metric_dict = metric_aggregator.average()

    msg_1 = "Epoch: {}, Validation on test samples:".format(epoch)
    msg_2 = f"RMSE: {metric_dict['rmse']:.3f} RMSE_log: {metric_dict['rmse_log']:.3f} REL: {metric_dict['absrel']:.3f} Sq_rel: {metric_dict['sq_rel']:.3f}"
    msg_3 = f"Delta1: {metric_dict['delta1']:.3f} Delta2: {metric_dict['delta2']:.3f} Delta3: {metric_dict['delta3']:.3f}" 
    print(msg_1)
    print(msg_2)
    print(msg_3)

    with open(log_file, 'a') as f:
        print(msg_1, file=f)
        print(msg_2, file=f)
        print(msg_3, file=f)

if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument("--config", required=True, type=str)
    args = parser.parse_args()
    config = yaml.load(open(args.config, "r"), Loader = yaml.FullLoader)

    #random.seed(config["random_seed"])
    #np.random.seed(config["random_seed"])
    torch.manual_seed(config['random_seed'])
    torch.cuda.manual_seed(config["random_seed"])

    device = config['device']

    print("Device: {}".format(device))

    data_path = config['data_path']

    batch_size = config['batch_size']

    test_flag = config['test_flag']
    workers = config['workers']

    with open(config['test_filepath'], 'r') as f:
        test_filenames = f.read().splitlines()

    # create test frame dict
    test_frames_dict = sort_test_seq_func(test_filenames) 
    
    if test_flag:
        print("Only validating on test sequences:")       

        dim_dict = {
            'input_dim':512,
            'width': config['width'] // 32,
            'height':config['height'] // 32,
            'hidden_dim_1':512,
            'hidden_dim_2':512
        }
        kernel_size = (3,3)
        # initialize encoder
        encoder = ResnetEncoder(18, config['pretrained_enc'])
        
        exp_var = config['exp_v']

        # initialize decoder/ clstm
        if exp_var == 'v1':
            clstm_model = ConvLSTMModel_v1(dim_dict, kernel_size, True, device)
            decoder = DepthDecoder_v1(num_ch_enc=encoder.num_ch_enc)
        elif exp_var == 'v2':
            clstm_model = ConvLSTMModel_v2(dim_dict, kernel_size, True, device)
            decoder = DepthDecoder_v1(num_ch_enc=encoder.num_ch_enc)
        elif exp_var == 'v3':
            clstm_model = ConvGRUModel_v1(dim_dict, kernel_size, True, device)
            decoder = DepthDecoder_v1(num_ch_enc=encoder.num_ch_enc)
        elif exp_var == 'v4':
            if config['v4_type'] == 'concat':
                dim_dict['hidden_dim_1'] = 256

            clstm_model = ConvGRUModel_v1(dim_dict, kernel_size, True, device)
            decoder = DepthDecoder_v1(num_ch_enc=encoder.num_ch_enc)
        
        if not config['debug']:
            pass
            # # load state dicts for encoder
            # loaded_dict_enc = torch.load(config['encoder_model_path'], map_location=device)
            # filtered_dict_enc = {k: v for k, v in loaded_dict_enc.items() if k in encoder.state_dict()}
            # encoder.load_state_dict(filtered_dict_enc)

            # # load state dicts for decoder 
            # loaded_dict = torch.load(config['decoder_model_path'], map_location=device)
            # decoder.load_state_dict(loaded_dict)

            # # load state dict for clstm

            # loaded_dict = torch.load(config['clstm_model_path'], map_location=device)
            # clstm_model.load_state_dict(loaded_dict)

            # print("Successfully loaded state dicts")

        encoder = encoder.to(device)
        decoder = decoder.to(device)
        clstm_model = clstm_model.to(device)

        test_validate_v1(test_frames_dict, encoder, clstm_model, decoder, config)
    else:
        
        if config['use_rnn']:
            train_seq_filepath = config['train_seq_path']
            val_seq_filepath = config['val_seq_path'] 

            with open(train_seq_filepath, 'r') as f:
                train_seq = f.read().splitlines()
                
            with open(val_seq_filepath, 'r') as f:
                val_seq = f.read().splitlines()

            train_n_frames_dict = count_scene_frames_func(train_seq, data_path)
            val_n_frames_dict = count_scene_frames_func(val_seq, data_path)

            n = config['n_sequence']
            k = config['att_n_mask']

            train_tuples = generate_frame_seq_func(train_n_frames_dict, train_seq, n, k, config['train_n_tuples'])
            val_tuples = generate_frame_seq_func(val_n_frames_dict, val_seq, n, k, config['test_n_tuples'])

            print('Generated train tuples: {}, val tuples: {}'.format(len(train_tuples), len(val_tuples)))
            train_dataset = KITTIDataset_v1(config, train_tuples, True)
            val_dataset = KITTIDataset_v1(config, val_tuples, False)

        else:
            with open(config['train_filepath'], 'r') as f:
                train_filenames = f.read().splitlines()
            with open(config['test_filepath'], 'r') as f:
                test_filenames = f.read().splitlines()

            train_dataset = KITTIDataset_v2(config, train_filenames, True)
            val_dataset = KITTIDataset_v2(config, test_filenames, False)   
        

        print("Initialized datasets")    
        
        # 2. initialize dataloaders
        train_loader = DataLoader(train_dataset, batch_size, True, num_workers=workers, pin_memory=True, drop_last=True)
        val_loader = DataLoader(val_dataset, batch_size, False, num_workers=workers, pin_memory=True, drop_last=False)

        # 3. initialize models (encoder, rnn, decoder)
        use_skips = config['use_skips']

        encoder = ResnetEncoder(18, config['pretrained_enc'])
        encoder = encoder.to(device)

        if config['use_rnn']:

            dim_dict = {
            'input_dim':encoder.num_ch_enc[-1],
            'width': config['width'] // 32,
            'height':config['height'] // 32,
            'hidden_dim_1':512,
            'hidden_dim_2':512
            }

            k_s = config['kernel_size']
            kernel_size = (k_s, k_s)

            exp_var = config['exp_v']

            if exp_var == 'v1':
                # just lstm
                clstm_model = ConvLSTMModel_v1(dim_dict, kernel_size, True, device)
                decoder = DepthDecoder_v1(num_ch_enc=encoder.num_ch_enc, use_skips=use_skips)
            elif exp_var == 'v2':
                # lstm with pseudo attention
                dim_dict['hidden_dim_1'] = 256
                dim_dict['hidden_dim_2'] = 256
                clstm_model = ConvLSTMModel_v2(dim_dict, kernel_size, True, device)
                decoder = DepthDecoder_v1(num_ch_enc=encoder.num_ch_enc, use_skips=use_skips)
            elif exp_var == 'v3':
                # just gru
                clstm_model = ConvGRUModel_v1(dim_dict, kernel_size, True, device)
                decoder = DepthDecoder_v1(num_ch_enc=encoder.num_ch_enc)
            elif exp_var == 'v4':
                # gru + fusion/concat of previous hidden state
                if config['v4_type'] == 'concat':
                    dim_dict['hidden_dim_1'] = 256
                clstm_model = ConvGRUModel_v1(dim_dict, kernel_size, True, device)
                decoder = DepthDecoder_v1(num_ch_enc=encoder.num_ch_enc)
            elif exp_var == 'v5':
                # lstm + fusion/concat of previous hidden state
                if config['v5_type'] == 'concat':
                    dim_dict['hidden_dim_1'] = 256
                clstm_model = ConvLSTMModel_v1(dim_dict, kernel_size, True, device)
                decoder = DepthDecoder_v1(num_ch_enc=encoder.num_ch_enc)
            elif exp_var == 'v6':
                dim_dict['hidden_dim_1'] = 256
                dim_dict['hidden_dim_2'] = 256
                clstm_model = ConvGRUModel_v2(dim_dict, kernel_size, True, device)
                decoder = DepthDecoder_v1(num_ch_enc=encoder.num_ch_enc, use_skips=use_skips)                

            decoder = decoder.to(device)
            clstm_model = clstm_model.to(device)

            # Test this
            # load state dicts for encoder
            if config['use_pretrained_enc_dec']:
                loaded_dict_enc = torch.load(config['encoder_best_ss_model'], map_location=device)
                filtered_dict_enc = {k: v for k, v in loaded_dict_enc.items() if k in encoder.state_dict()}
                encoder.load_state_dict(filtered_dict_enc)

                # load state dicts for decoder 
                loaded_dict = torch.load(config['decoder_best_ss_model'], map_location=device)
                decoder.load_state_dict(loaded_dict)

                print("Loaded pretrained enc/dec states")
            else:
                print("Haven't load pretrained enc/dec states")

            parameters = list(encoder.parameters()) + list(decoder.parameters()) + list(clstm_model.parameters())

            h_s_epoch = config['h_s_epoch']

        else:
            decoder = DepthDecoder_v1(num_ch_enc=encoder.num_ch_enc, use_skips=use_skips)
            decoder = decoder.to(device)
            parameters = list(encoder.parameters()) + list(decoder.parameters())


        # 4. initialized criterions 
        criterion_name = config['criterion']

        if criterion_name == 'l2':
            criterion = MaskedMSELoss().to(device)
        elif criterion_name == 'l1':
            criterion = MaskedL1Loss().to(device)
        elif criterion_name == 'berhu':
            criterion = BerHu().to(device)

        print("Initialize criterion")

        # 5. initialize optimizers
        opt_name = config['optimizer']
        opt_params = config['opt_params']
        # opt_params example {'learning_rate':0.01, 'weight_decay':1e-4, 'momentum': 0.9}
        
        if opt_name == 'adam':
            optimizer = torch.optim.Adam(parameters, **opt_params)
        elif opt_name == 'sgd':
            optimizer = torch.optim.SGD(parameters, **opt_params)

        lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer=optimizer, gamma=config['lr_decay_rate'])

        print("Initialized optimizer")
        
        # initialize folders to store models, etc    

        current_date = datetime.datetime.now().strftime('%d%m%y')
        current_time = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
        checkpoint_path = f"{config['checkpoint_path']}/{current_date}/{config['run_name']}_{current_time}" 
        encoder_model_path = os.path.join(checkpoint_path, "encoder")
        decoder_model_path = os.path.join(checkpoint_path, "decoder")

        if config['use_rnn']:
            clstm_model_path = os.path.join(checkpoint_path, "clstm")

        if not os.path.exists(checkpoint_path):
            os.makedirs(checkpoint_path)

        if not os.path.exists(encoder_model_path):
            os.makedirs(encoder_model_path)
        
        if not os.path.exists(decoder_model_path):
            os.makedirs(decoder_model_path)

        if config['use_rnn'] and not os.path.exists(clstm_model_path):
            os.makedirs(clstm_model_path)

        print("Initialized store folders")

        logfile_path = os.path.join(checkpoint_path, 'log.txt')

        with open(logfile_path, 'w') as f:
            print("", file=f)

        copy2(args.config, checkpoint_path)

        num_epochs = config['n_epochs']

        print("Validate before training:")

        if config['use_rnn']:
            encoder.eval()
            #validate_rnn(val_loader, encoder, decoder, clstm_model, device, 0, logfile_path)
            if exp_var in ['v1', 'v3', 'v4', 'v5']:
                test_validate_v1(test_frames_dict, encoder, clstm_model, decoder, config, logfile_path)
            elif exp_var == 'v2':
                test_validate_v2(test_frames_dict, encoder, clstm_model, decoder, config, logfile_path)
            elif exp_var == 'v6':
                test_validate_v3(test_frames_dict, encoder, clstm_model, decoder, config, logfile_path)

        else:
            encoder.eval()
            validate(val_loader, encoder, decoder, device, 0, logfile_path)

        print("Start Training:")

        for epoch in range(num_epochs):
            print("Epoch: {}".format(epoch+1))
            encoder.train()
            decoder.train()

            if config['use_rnn']:
                clstm_model.train()

            if config['use_rnn'] and (epoch + 1) == h_s_epoch:
                print("Removed hidden states from optimizer")
                clstm_model.h0_layer1.requires_grad = False
                if exp_var in ['v1','v5']:
                    clstm_model.c0_layer1.requires_grad = False
                if exp_var == 'v2':
                    clstm_model.c0_layer1.requires_grad = False
                    clstm_model.h0_layer2.requires_grad = False
                    clstm_model.c0_layer2.requires_grad = False
                if exp_var == 'v6':
                    clstm_model.h0_layer2.requires_grad = False

            if config['use_rnn']:
                train_epoch_rnn(train_loader, encoder, decoder, clstm_model, optimizer, epoch, device, config, logfile_path)
            else:
                train_epoch(train_loader, encoder, decoder, optimizer, epoch, device, config, logfile_path)
            
            print("Validating:")
            encoder.eval()
            decoder.eval()

            if config['use_rnn']:
                clstm_model.eval()
                if exp_var in ['v1', 'v3', 'v4', 'v5']:
                    test_validate_v1(test_frames_dict, encoder, clstm_model, decoder, config, logfile_path)
                elif exp_var == 'v2':
                    test_validate_v2(test_frames_dict, encoder, clstm_model, decoder, config, logfile_path)
                elif exp_var == 'v6':
                    test_validate_v3(test_frames_dict, encoder, clstm_model, decoder, config, logfile_path)
            else:
                validate(val_loader, encoder, decoder, device, epoch+1, logfile_pathv

            lr_scheduler.step()

            torch.save(encoder.state_dict(), f"{encoder_model_path}/encoder_model_{epoch+1}.pth")
            torch.save(decoder.state_dict(), f"{decoder_model_path}/decoder_model_{epoch+1}.pth")

            if config['use_rnn']:
                torch.save(clstm_model.state_dict(), f"{clstm_model_path}/clstm_model_{epoch+1}.pth")

            print("Successfully saved models")
            print("-------------------------")
