import torch
import torch.nn as nn

import math
import numpy as np
import torch.nn.functional as F

class ConvLSTMCell_v1(nn.Module):

    def __init__(self, input_dim, hidden_dim, kernel_size, bias):
        """
        Initialize ConvLSTM cell.
        Parameters
        ----------
        input_dim: int
            Number of channels of input tensor.
        hidden_dim: int
            Number of channels of hidden state.
        kernel_size: (int, int)
            Size of the convolutional kernel.
        bias: bool
            Whether or not to add the bias.
        """

        super(ConvLSTMCell_v1, self).__init__()

        self.input_dim = input_dim
        self.hidden_dim = hidden_dim

        self.kernel_size = kernel_size
        self.padding = kernel_size[0] // 2, kernel_size[1] // 2
        self.bias = bias

        self.conv = nn.Conv2d(in_channels=self.input_dim + self.hidden_dim,
                              out_channels=4 * self.hidden_dim,
                              kernel_size=self.kernel_size,
                              padding=self.padding,
                              bias=self.bias)

    def forward(self, input_tensor, cur_state):
        h_cur, c_cur = cur_state

        combined = torch.cat([input_tensor, h_cur], dim=1)  # concatenate along channel axis

        combined_conv = self.conv(combined)
        cc_i, cc_f, cc_o, cc_g = torch.split(combined_conv, self.hidden_dim, dim=1)
        i = torch.sigmoid(cc_i)
        f = torch.sigmoid(cc_f)
        o = torch.sigmoid(cc_o)
        g = torch.tanh(cc_g)

        c_next = f * c_cur + i * g
        h_next = o * torch.tanh(c_next)

        return h_next, c_next

class ConvLSTMCell_v2(nn.Module):

    def __init__(self, input_dim, hidden_dim, height, width, kernel_size, bias):
        """
        Initialize ConvLSTM cell.
        Parameters
        ----------
        input_dim: int
            Number of channels of input tensor.
        hidden_dim: int
            Number of channels of hidden state.
        kernel_size: (int, int)
            Size of the convolutional kernel.
        bias: bool
            Whether or not to add the bias.
        """

        super(ConvLSTMCell_v2, self).__init__()

        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.height = height
        self.width = width
        self.kernel_size = kernel_size
        self.padding = kernel_size[0] // 2, kernel_size[1] // 2
        self.bias = bias

        self.conv = nn.Conv2d(in_channels=self.input_dim + self.hidden_dim,
                              out_channels=4 * self.hidden_dim,
                              kernel_size=self.kernel_size,
                              padding=self.padding,
                              bias=self.bias)

        self.Wci = nn.Parameter(torch.zeros(1, self.hidden_dim, self.height, self.width), requires_grad=True)
        self.Wcf = nn.Parameter(torch.zeros(1, self.hidden_dim, self.height, self.width), requires_grad=True)
        self.Wco = nn.Parameter(torch.zeros(1, self.hidden_dim, self.height, self.width), requires_grad=True)

    def forward(self, input_tensor, cur_state):
        h_cur, c_cur = cur_state
        #print(input_tensor.shape, h_cur.shape)
        combined = torch.cat([input_tensor, h_cur], dim=1)  # concatenate along channel axis

        combined_conv = self.conv(combined)
        cc_i, cc_f, cc_g, cc_o = torch.split(combined_conv, self.hidden_dim, dim=1)
        i = torch.sigmoid(cc_i + self.Wci* c_cur)
        f = torch.sigmoid(cc_f + self.Wcf* c_cur)
        o = torch.sigmoid(cc_o + self.Wco* c_cur)

        c_next = f * c_cur + i * F.elu(cc_g)
        h_next = o * F.elu(c_next)

        return h_next, c_next

class ConvLSTMModel_v1(nn.Module):
    def __init__(self, dim_dict, kernel_size, bias, device):
        super(ConvLSTMModel_v1, self).__init__()

        self.clstm_1 = ConvLSTMCell_v1(dim_dict['input_dim'], dim_dict['hidden_dim_1'], kernel_size, bias)
        self.h0_layer1, self.c0_layer1 = self.init_hidden(dim_dict['hidden_dim_1'], dim_dict['height'], dim_dict['width'], device)
        self.height = dim_dict['height']
        self.width = dim_dict['width']
        self.hidden_dim = dim_dict['hidden_dim_1']

    def init_hidden(self, hidden_dim, height, width, device):
        return (nn.Parameter(torch.zeros(1, hidden_dim, height, width, device=device), requires_grad=True),
                nn.Parameter(torch.zeros(1, hidden_dim, height, width, device=device), requires_grad=True))

    def forward(self, x, hidden_state_1):
        return self.clstm_1(x, hidden_state_1)

class ConvLSTMModel_v2(nn.Module):
    def __init__(self, dim_dict, kernel_size, bias, device):
        super(ConvLSTMModel_v2, self).__init__()

        self.clstm_1 = ConvLSTMCell_v1(dim_dict['input_dim'], dim_dict['hidden_dim_1'], kernel_size, bias)
        self.clstm_2 = ConvLSTMCell_v1(dim_dict['hidden_dim_1'], dim_dict['hidden_dim_2'], kernel_size, bias)

        self.h0_layer1, self.c0_layer1 = self.init_hidden(dim_dict['hidden_dim_1'], dim_dict['height'], dim_dict['width'], device)
        self.h0_layer2, self.c0_layer2 = self.init_hidden(dim_dict['hidden_dim_2'], dim_dict['height'], dim_dict['width'], device)


    def init_hidden(self, hidden_dim, height, width, device):
        return (nn.Parameter(torch.zeros(1, hidden_dim, height, width, device=device), requires_grad=True),
                nn.Parameter(torch.zeros(1, hidden_dim, height, width, device=device), requires_grad=True))

    def forward(self, x, hidden_state_1, hidden_state_2):

        hidden_state_1 = self.clstm_1(x, hidden_state_1)
        hidden_state_2 = self.clstm_2(hidden_state_1[0], hidden_state_2)

        return hidden_state_1, hidden_state_2

class ConvGRUCell(nn.Module):
    def __init__(self, input_size, input_dim, hidden_dim, kernel_size, bias):
        super(ConvGRUCell, self).__init__()

        self.height, self.width = input_size
        self.padding = kernel_size[0] // 2, kernel_size[1] // 2
        self.hidden_dim = hidden_dim
        self.bias = bias
        self.conv_gates = nn.Conv2d(in_channels=input_dim + hidden_dim,
                                    out_channels=2*self.hidden_dim,  # for update_gate,reset_gate respectively
                                    kernel_size=kernel_size,
                                    padding=self.padding,
                                    bias=self.bias)

        self.conv_can = nn.Conv2d(in_channels=input_dim+hidden_dim,
                              out_channels=self.hidden_dim, # for candidate neural memory
                              kernel_size=kernel_size,
                              padding=self.padding,
                              bias=self.bias)

    def init_hidden(self, batch_size):
        return torch.zeros(batch_size, self.hidden_dim, self.height, self.width)

    def forward(self, input_tensor, h_cur):
        combined = torch.cat([input_tensor, h_cur], dim=1)
        combined_conv = self.conv_gates(combined)

        gamma, beta = torch.split(combined_conv, self.hidden_dim, dim=1)
        reset_gate = torch.sigmoid(gamma)
        update_gate = torch.sigmoid(beta)

        combined = torch.cat([input_tensor, reset_gate*h_cur], dim=1)
        cc_cnm = self.conv_can(combined)
        #cnm = torch.tanh(cc_cnm)
        cnm = F.elu(cc_cnm)
        h_next = (1 - update_gate) * h_cur + update_gate * cnm
        
        return h_next

class ConvGRUModel_v1(nn.Module):
    def __init__(self, dim_dict, kernel_size, bias, device):
        super(ConvGRUModel_v1, self).__init__()

        self.cgru_1 = ConvGRUCell((dim_dict['height'], dim_dict['width']), dim_dict['input_dim'], dim_dict['hidden_dim_1'], kernel_size, bias)
        self.h0_layer1 = self.init_hidden(dim_dict['hidden_dim_1'], dim_dict['height'], dim_dict['width'], device)
        self.height = dim_dict['height']
        self.width = dim_dict['width']
        self.hidden_dim = dim_dict['hidden_dim_1']

    def init_hidden(self, hidden_dim, height, width, device):
        return nn.Parameter(torch.zeros(1, hidden_dim, height, width, device=device), requires_grad=True)

    def forward(self, x, hidden_state_1):
        return self.cgru_1(x, hidden_state_1)

class ConvGRUModel_v2(nn.Module):
    def __init__(self, dim_dict, kernel_size, bias, device):
        super(ConvGRUModel_v2, self).__init__()

        self.cgru_1 = ConvGRUCell((dim_dict['height'], dim_dict['width']), dim_dict['input_dim'], dim_dict['hidden_dim_1'], kernel_size, bias)
        self.cgru_2 = ConvGRUCell((dim_dict['height'], dim_dict['width']), dim_dict['hidden_dim_1'], dim_dict['hidden_dim_2'], kernel_size, bias)
        self.h0_layer1 = self.init_hidden(dim_dict['hidden_dim_1'], dim_dict['height'], dim_dict['width'], device)
        self.h0_layer2 = self.init_hidden(dim_dict['hidden_dim_2'], dim_dict['height'], dim_dict['width'], device)
        self.height = dim_dict['height']
        self.width = dim_dict['width']
        self.hidden_dim = dim_dict['hidden_dim_1']

    def init_hidden(self, hidden_dim, height, width, device):
        return nn.Parameter(torch.zeros(1, hidden_dim, height, width, device=device), requires_grad=True)

    def forward(self, x, hidden_state_1, hidden_state_2):
        hidden_state_1 = self.cgru_1(x, hidden_state_1)
        hidden_state_2 = self.cgru_2(hidden_state_1, hidden_state_2)
        
        return hidden_state_1, hidden_state_2